/*
 * EfficientSPC -- Implementation
 *
 * Author:  Henrik Zimmer <henrik@zimmer.to>
 *
 * Version: $Revision: 1 $
 * Date:    $date$
 */


#include "EfficientSPC.hh"

////////////////////////////////////////////////////////

namespace ACG {
  
////////////////////////////////////////////////////////

Scalar EfficientSPC::tetrahedralize( MeshT& _mesh, PMeshT& _robusthull, Scalar _avgel, LeanTetMesh& _tetmesh, Scalar _edgescale, Scalar _aspect)
{
  ACG::StopWatch sw;
  sw.start();

  int nv( _mesh.n_vertices());
  int nf( _mesh.n_faces());
  int ne( _mesh.n_edges());

  std::vector< int > movedvhs( nv, -1);

  tetgenio in, out;
  tetgenio::facet *f;
  tetgenio::polygon *p;
  int i;

  ////! setup EfficientSPC Structures
  in.firstnumber = 0;
  in.numberofpoints = nv + _robusthull.n_vertices();
  in.pointlist = new REAL[in.numberofpoints * 3];
  // NOTE! The vertices of the original mesh [0,nv-1] are the 
  // first ones, this indexing is kept by EfficientSPC, important later

  // add vertices of original mesh
  for( VIter v_it = _mesh.vertices_begin(); v_it != _mesh.vertices_end(); ++v_it)
  {
    int vid( v_it.handle().idx());
    int id( vid * 3);
    Vec3 pt( _mesh.point(v_it));
    in.pointlist[id+0] = pt[0];
    in.pointlist[id+1] = pt[1];
    in.pointlist[id+2] = pt[2];
  }

  // add vertices on robust hull
  for( VIter v_it = _robusthull.vertices_begin(); v_it != _robusthull.vertices_end(); ++v_it)
  {
    int vid( v_it.handle().idx()+nv);
    int id( vid * 3);
    Vec3 pt( _robusthull.point(v_it));
    in.pointlist[id+0] = pt[0];
    in.pointlist[id+1] = pt[1];
    in.pointlist[id+2] = pt[2];
  }


  // add faces of the meshes to the EfficientSPC structure, both inward
  // pointing, i.e., in different orientations
  in.numberoffacets = nf+_robusthull.n_faces();
  in.facetlist = new tetgenio::facet[in.numberoffacets];
  in.facetmarkerlist = new int[in.numberoffacets]; // face tags

  // add faces of original mesh (in original orientation)
  for( FaceIter f_it = _mesh.faces_begin(); f_it != _mesh.faces_end(); ++f_it)
  {
    int fid( f_it.handle().idx());
    in.facetmarkerlist[fid] = ORIGMESHTAG;

    f = &in.facetlist[fid];
    f->numberofpolygons = 1;
    f->polygonlist = new tetgenio::polygon[f->numberofpolygons];
    f->numberofholes = 0;
    f->holelist = NULL;
    p = &f->polygonlist[0];
    p->numberofvertices = 3; // triangles
    p->vertexlist = new int[p->numberofvertices];
    int vcnt(0);
    for( FVIter fv_it = _mesh.fv_iter( f_it); fv_it; ++fv_it)
    {
      int vid( fv_it.handle().idx());
      p->vertexlist[p->numberofvertices-vcnt-1] = vid;
      vcnt++;
    }
  }

  // add faces of robust hull mesh in opposite orientation (polygon mesh!)
  for( PFIter f_it = _robusthull.faces_begin(); f_it != _robusthull.faces_end(); ++f_it)
  {
    int fid( f_it.handle().idx() + nf);
    in.facetmarkerlist[fid] = 33;

    f = &in.facetlist[fid];
    f->numberofpolygons = 1;
    f->polygonlist = new tetgenio::polygon[f->numberofpolygons];
    f->numberofholes = 0;
    f->holelist = NULL;
    p = &f->polygonlist[0];
    p->numberofvertices = _robusthull.valence(f_it.handle());
    p->vertexlist = new int[p->numberofvertices];
    int vcnt(0);
    for( PFVIter fv_it = _robusthull.fv_iter( f_it); fv_it; ++fv_it)
    {
      int vid( fv_it.handle().idx()+nv);
      p->vertexlist[p->numberofvertices-vcnt-1] = vid; // reverse orientation
      vcnt++;
    }
  }

  // set "inside point" to carve out (not tetrahedralize) interior of original mesh to keep element count as low as possible
  Vec3 insidept( _mesh.point( VH(0)) - _mesh.normal( VH(0))*_avgel*0.0001);
  // NOTE! Shifting works well in practise, for a guarantee first quickly tetrahedralize interior of original mesh (without quality constraints) and use COG of some cell
  in.numberofholes = 1;
  in.holelist = new REAL[in.numberofholes];
  in.holelist[0] = insidept[0];
  in.holelist[1] = insidept[1];
  in.holelist[2] = insidept[2];

  ////! Quality TetMesh Generation
  // do quality mesh generation (q) with specified quality bounds
  // aspect ratio and maxvolume
  double maxvol( sqrt(512.)/192.*_avgel*_avgel*_avgel);
  double scalevol(_edgescale);
  QString maxvolstr(QString::number( maxvol*scalevol*scalevol*scalevol, 'f', 9));
  // Maxvolume is volume of a uniform tetrahedron with _edgescale*_avgel sides
  std::string options("");
  options += "p";
  options += "q" + QString::number( _aspect, 'f',4).toStdString();
  options += "Q"; // quiet
  options += "a"+ maxvolstr.toStdString();
  options += "z";
  char * optionscstr = new char [options.size()+1];
  strcpy (optionscstr, options.c_str());
  std::cerr << options << std::endl;
  try
  {
    ::tetrahedralize(optionscstr, &in, &out);

    size_t ntetv( out.numberofpoints); 
    size_t ntetc( out.numberoftetrahedra);
    size_t ntetf( out.numberoftrifaces); // if flag f?!
    std::cerr << " CDT #tets  = " << ntetc << std::endl;
    std::cerr << " CDT #pts   = " << ntetv << std::endl;
    std::cerr << " CDT #tri   = " << ntetf << std::endl;

    // use own small footprint mesh structure
    _tetmesh = LeanTetMesh( ntetv, ntetc);
    // setup points
    for( size_t i = 0; i < ntetv; ++i)
    {
      int id(i*3);
      _tetmesh.pts[i] = Vec3(out.pointlist[id], out.pointlist[id+1], out.pointlist[id+2]);
    }

    // tag boudnary vertices
    for( size_t i = 0; i < ntetf; ++i)
    {
      int id(i*3);
      if( out.trifacemarkerlist[i] == ROBUSTHULLTAG)
      {
        _tetmesh.tag[out.trifacelist[id]] = ROBUSTHULLTAG;
        _tetmesh.tag[out.trifacelist[id+1]] = ROBUSTHULLTAG;
        _tetmesh.tag[out.trifacelist[id+2]] = ROBUSTHULLTAG;
      }
    }

    std::cerr << " ... setting up tet mesh ... " << std::endl;

    int* tetlist( out.tetrahedronlist);
    for( size_t i = 0; i < ntetc; ++i)
    {
      int id(4*i);

      for( size_t v = 0; v < 4; ++v)
      {
        _tetmesh.tets[i][v] = tetlist[id+v];
        // map: vertex -> cells
        _tetmesh.vc[tetlist[id+v]].insert( i);
      }

      for( size_t v = 0; v < 4; ++v)
      {
        int vidf(tetlist[id+v]);
        for( size_t o = 0; o < 4; ++o)
        {
          if( o != v)
          {
            int vido( tetlist[id+o]);
            // map: vertex -> vertex neighbors
            _tetmesh.vv[vidf].insert( vido);
            // NOTE! std::set guarantees that each neighbor
            // only appears once in vv
          }
        }
      }
    }
    _tetmesh.build_edges();

  }
  catch(...)
  {
    std::cerr << __FUNCTION__ << " EfficientSPC Failed! possibly self-intersecting faces in the input mesh... " << std::endl;
  }
  return sw.stop()/1000.;
}

////////////////////////////////////////////////////////

Scalar EfficientSPC::do_convex_hull_cgal( MeshT& _inpmesh, MeshT& _outmesh)
{
  /*
  ACG::StopWatch sw;
  sw.start();

  std::vector< Point_3 > points( _inpmesh.n_vertices());
  for( size_t p = 0; p < points.size(); ++p)
  {
    Vec3 pt( _inpmesh.point( VH(p)));
    points[p] = Point_3(pt[0], pt[1], pt[2]);
  }

  ACG::StopWatch sw;
  sw.start();

  // define object to hold convex hull
  CGAL::Object ch_object;

  // compute convex hull
  CGAL::convex_hull_3(points.begin(), points.end(), ch_object);

  Polyhedron_3 convexhull;

  // determine what kind of object it is
  if (CGAL::object_cast<Segment_3>(&ch_object) )
  {
    std::cout << "convex hull is a segment " << std::endl;
    return -1;
  }
  else if (CGAL::object_cast<Polyhedron_3>(&ch_object) )
  {
    std::cout << "convex hull is a polyhedron " << std::endl;
    convexhull = CGAL::object_cast<Polyhedron_3>(ch_object);
  }
  else
  {
    std::cout << "convex hull error!" << std::endl;
    return -1;
  }

  std::cerr << " convex hull took " << (sw.stop()/1000.0) << std::endl;


  std::vector< VH > face(3, VH(-1));

  size_t f(0);

  std::map< Polyhedron_3::Vertex_iterator, TriMesh::VertexHandle > vertmap;
  for( Polyhedron_3::Vertex_iterator vert_it = convexhull.vertices_begin(); vert_it != convexhull.vertices_end(); ++vert_it)
  {
    Point_3& curpt( vert_it->point() );
    vertmap[vert_it] = _outmesh.add_vertex( Vec3(curpt[0], curpt[1], curpt[2]));
  }
  std::cerr << "_outmesh.n_vertices()" << _outmesh.n_vertices() << std::endl;

  // mesh it
  size_t sizef = convexhull.size_of_facets();
  PolyFIter faceit = convexhull.facets_begin();
  PolyFIter facend = convexhull.facets_end();

  for(; faceit != facend; ++faceit )
  {
    face.clear();
    //PolyFHIter fhiter(faceit->facet_begin());
    if( !faceit->is_triangle() )
    {
      std::cerr << __FUNCTION__ << " NOT TRAINGLE !!! " << std::endl;
      return -1;
    }
    PolyFHIter fhiter = faceit->facet_begin();
    PolyFHIter startr = faceit->facet_begin();
    int cnt(0);
    do
    {
      //Point_3& curpt( fhiter->vertex()->point() );
      //face[cnt] = convmesh->add_vertex( TriMesh::Point(curpt[0], curpt[1], curpt[2] ) );
      face.push_back( vertmap[fhiter->vertex()]);
      cnt++;
      fhiter++;
    } while( startr != fhiter );
    std::vector< VH > revface(face.size());
    for( size_t j = 0; j < face.size(); ++j)
      revface[face.size()-j-1] = face[j];
    _outmesh.add_face( revface );
  }
  _outmesh.update_normals();
  return sw.stop()/1000.;
  */
    return -1;
}

////////////////////////////////////////////////////////

Scalar EfficientSPC::do_convex_hull_tetgen( MeshT& _inpmesh, MeshT& _outmesh)
{
  ACG::StopWatch sw;
  sw.start();

  const int nv( _inpmesh.n_vertices());
  const int nf( _inpmesh.n_faces());
  const int ne( _inpmesh.n_edges());

  /// compute average edgelength in input mesh
  Scalar el(0.0);
  for( TriMesh::EdgeIter e_it = _inpmesh.edges_begin(); e_it != _inpmesh.edges_end(); ++e_it)
    el+= (_inpmesh.point( _inpmesh.to_vertex_handle( _inpmesh.halfedge_handle( e_it, 0)))  - _inpmesh.point( _inpmesh.to_vertex_handle( _inpmesh.halfedge_handle( e_it,1)))).norm();
  el /= Scalar(ne);


  ////! setup EfficientSPC Structures
  tetgenio in, out;
  tetgenio::facet *f;
  tetgenio::polygon *p;
  int i;
  // All indices start from 0.
  in.firstnumber = 0;

  // setup input vertex positions
  in.numberofpoints = nv;
  in.pointlist = new REAL[in.numberofpoints * 3];

  for( TriMesh::VertexIter v_it = _inpmesh.vertices_begin(); v_it != _inpmesh.vertices_end(); ++v_it)
  {
    int vid( v_it.handle().idx());
    int id( vid * 3);
    const TriMesh::Point pt( _inpmesh.point(v_it));
    in.pointlist[id+0] = pt[0];
    in.pointlist[id+1] = pt[1];
    in.pointlist[id+2] = pt[2];
    //in.pointattributelist[vid] = vid;
  }

  // setup input faces
  in.numberoffacets = nf;
  in.facetlist = new tetgenio::facet[in.numberoffacets];
  in.facetmarkerlist = new int[in.numberoffacets];

  for( FaceIter f_it = _inpmesh.faces_begin(); f_it != _inpmesh.faces_end(); ++f_it)
  {
    int fid( f_it.handle().idx());

    f = &in.facetlist[fid];
    f->numberofpolygons = 1;
    f->polygonlist = new tetgenio::polygon[f->numberofpolygons];
    f->numberofholes = 0;
    f->holelist = NULL;
    p = &f->polygonlist[0];
    p->numberofvertices = 3; // assuming input triangle mesh
    p->vertexlist = new int[p->numberofvertices];
    int vcnt(0);
    for( TriMesh::FaceVertexIter fv_it = _inpmesh.fv_iter( f_it); fv_it; ++fv_it)
    {
      int vid( fv_it.handle().idx());
      p->vertexlist[vcnt] = vid;
      vcnt++;
    }
  }

  ////! do mesh generation (Convex Hull) 
  std::string options("");
  //options += "q2.42";
  options += "Q"; // quiet
  options += "z";
  char * optionscstr = new char [options.size()+1];
  strcpy (optionscstr, options.c_str());
  //std::cerr << options << std::endl;
  ::tetrahedralize(optionscstr, &in, &out);

  // number of vertices and faces of the convex hull
  size_t nconv( out.numberofpoints);
  size_t nconf( out.numberoftrifaces); 

  ////! extract triangle mesh of convex hull
  // add vertices (adjacent to faces) only once
  std::vector< int > vhmap( nconv, -1);
  for( size_t i = 0; i < nconf; ++i)
  {
    int id(3*i);
    for( size_t v = 0; v < 3; ++v)
    {
      if( vhmap[out.trifacelist[id+v]] == -1)
      {
        int vid( out.trifacelist[id+v] * 3);
        VH vh(_outmesh.add_vertex( Vec3(out.pointlist[vid], out.pointlist[vid+1], out.pointlist[vid+2])));
        vhmap[out.trifacelist[id+v]] = vh.idx();
      }
    }
  }

  // add faces
  std::vector< VH > facevhs;
  facevhs.reserve(3);

  for( size_t i = 0; i < nconf; ++i)
  {
    int id(3*i);
    facevhs.clear();
    facevhs.push_back( _outmesh.vertex_handle( vhmap[out.trifacelist[id+0]]));
    facevhs.push_back( _outmesh.vertex_handle( vhmap[out.trifacelist[id+1]]));
    facevhs.push_back( _outmesh.vertex_handle( vhmap[out.trifacelist[id+2]]));
    _outmesh.add_face( facevhs);
  }

  _outmesh.update_normals();

  ////! compute stable normals of convex hull faces
  // the convex hull is a very irregular tessellation. A simple
  // average can yield unsatisfactory results for normals of
  // vertices having many indicent faces on "one side".
  // Weight by angles to get more stable normal.
  for( VIter v_it = _outmesh.vertices_begin(); v_it != _outmesh.vertices_end(); ++v_it)
  {
    const Vec3 origpoint( _outmesh.point( v_it));
    VOHIter voh_it( _outmesh.voh_iter( v_it));
    Scalar anglesum(0.0);
    Vec3 averagednormal(0.0,0.0,0.0);
    for( ; voh_it; ++voh_it)
    {
      Vec3 vec1( _outmesh.point(_outmesh.to_vertex_handle( voh_it.handle())) - origpoint);
      Vec3 vec2( _outmesh.point(_outmesh.to_vertex_handle( _outmesh.next_halfedge_handle(voh_it.handle()))) - origpoint);
      vec1.normalize();
      vec2.normalize();
      Scalar angle( acos( std::min( 1.0, std::max( -1.0, vec1|vec2))));
      anglesum += angle;
      averagednormal += _outmesh.normal( _outmesh.face_handle( voh_it.handle())) * angle;
    }
    averagednormal /= anglesum;
    averagednormal.normalize();
    _outmesh.set_normal( v_it, averagednormal);
  }

  return sw.stop()/1000.;
}

////////////////////////////////////////////////////////

void EfficientSPC::fmm( LeanTetMesh& _tetmesh, MeshT& _hull, OpenMeshTriangleBSPT< MeshT >*& _chbsp, Scalar _avgel, MeshT& _mesh, Scalar& _inttime, Scalar& _dijtime)
{
  int nv( _mesh.n_vertices());

  ACG::StopWatch sw;
  sw.start();
  const int DEAD = 0; const int CLOSE = 1; const int FAR = 2; const int OUTSIDE = ROBUSTHULLTAG;
  std::vector< Scalar > fmm_dst( _tetmesh.nv, FLT_MAX);
  std::vector< int >    fmm_tag( _tetmesh.nv, FAR);

  std::set< std::pair< Scalar, int > > pq;

  ////! Initialize distances

  // Initialize distances of outer vertices to convex hull
  // use bsp of convex hull for efficiency
  for( size_t v = 0; v < _tetmesh.nv; ++v)
  {
    if( _tetmesh.tag[v] == OUTSIDE)
    {
      // compute closest distance to hull
      {
        FH bspfh( closest_bsp_fh( _tetmesh.pts[v], _chbsp));
        FVIter fv_it( _hull.fv_iter( bspfh));
        Vec3 hullpt( _hull.point( fv_it));
        Vec3 hullnorm( _hull.normal( bspfh));
        Scalar dist( ( _tetmesh.pts[v] - hullpt) | hullnorm);

        fmm_dst[v] = dist;
        fmm_tag[v] = DEAD;
      }

      // also add one ring to queue (CLOSE) // not necessary
      for( std::set<int>::iterator vvit( _tetmesh.vv[v].begin()); vvit != _tetmesh.vv[v].end(); ++vvit)
      {
        // dont add outside vertices to queue
        // and dont update already enqueued
        int vvid( *vvit);
        if( _tetmesh.tag[vvid] != OUTSIDE && fmm_tag[vvid] != CLOSE) 
        {
          FH bspfh( closest_bsp_fh( _tetmesh.pts[vvid], _chbsp));
          FVIter fv_it( _hull.fv_iter( bspfh));
          Vec3 hullpt( _hull.point( fv_it));
          Vec3 hullnorm( _hull.normal( bspfh));
          Scalar dist( ( _tetmesh.pts[vvid] - hullpt) | hullnorm);
          fmm_dst[vvid] = dist;
          fmm_tag[vvid] = CLOSE;
          pq.insert( std::make_pair( fmm_dst[vvid], vvid));
        }
      }
      //pq.insert( std::make_pair( fmm_dst[v], v));
    }
  }
  
  ////! Do FMM 
  /* See 
   *  "Computing first-arrival seismic traveltimes on 
   *  unstructured 3-D tetrahedral grids using the Fast 
   *  Marching Method" by Peter G. Leli`vre, 
   *  Colin G. Farquharson and Charles A. Hurich, 2010
   * for notation, e.g., vertex "D" etc.
   */

  while( !pq.empty())
  {
    // pop pair
    std::pair< Scalar, int > poppair( *pq.begin());
    Scalar popdist( poppair.first);
    int    popvid ( poppair.second);

    // erase entry
    pq.erase( pq.begin());

    //if( fmm_tag[popvid] != CLOSE)
    //  std::cerr << __FUNCTION__ << " ERROR should only pop CLOSE! ... is " << fmm_tag[popvid] << std::endl;

    // tag dead
    fmm_tag[popvid] = DEAD;

    // visit non-dead neighbors  (VertexVertexIter)
    for( std::set<int>::iterator vvit( _tetmesh.vv[popvid].begin()); vvit != _tetmesh.vv[popvid].end(); ++vvit)
    {
      int vvtovid( *vvit);

      // if DEAD don't visit
      if( fmm_tag[vvtovid] == DEAD)
        continue;

      // if not CLOSE and not FAR -> problem
      if( fmm_tag[vvtovid] != CLOSE && fmm_tag[vvtovid] != FAR)
        std::cerr << __FUNCTION__ << " PROBLEM, vertex should be CLOSE or FAR" << std::endl;

      // already in queue (CLOSE) then first remove
      if( fmm_tag[vvtovid] == CLOSE)
      {
        pq.erase( std::make_pair( fmm_dst[vvtovid], vvtovid));
      }

      /// compute new distance over all adjacent cells
      for( std::set<int>::iterator vcit( _tetmesh.vc[vvtovid].begin()); vcit != _tetmesh.vc[vvtovid].end(); ++vcit)
      {
        // do not waste time on cells which are do not contain any DEAD vertices
        bool cellok(false);

        // 1) sort points s.t. vertex "D" to be updated is the last one
        int offset(0);
        for( size_t cv = 0; cv < 4; ++cv)
        { 
          int cvid( _tetmesh.tets[*vcit][cv]);
          if( fmm_tag[cvid] == DEAD)
            cellok = true;

          if( cvid == vvtovid) // this is the vertex we want to update
            offset = cv;
        }
        if( !cellok)
          continue;

        // NOTE! Normals are needed by the FMM to project
        //       points, but the orientation depends on 
        //       the order of points, i.e., which of the
        //       cell vertices is currently "D"
        bool swaporientation(false);
        if( offset == 0 || offset == 2)
          swaporientation = false;
        else
          swaporientation = true;

        offset = 3 - offset;
        int vids[4]; // ids
        Vec3 pts[4]; // points
        Scalar t[4]; // distances (travel times)

        for( size_t cv = 0; cv < 4; ++cv)
        { 
          int curid( (cv+offset)%4);
          int oriid( _tetmesh.tets[*vcit][cv]);
          vids[curid] = oriid;
           pts[curid] = _tetmesh.pts[oriid];
             t[curid] = fmm_dst[oriid];
        }

        if( !swaporientation )
        {
          std::swap( vids[0], vids[2]);
          std::swap( pts[0], pts[2]);
          std::swap( t[0], t[2]);
        }

        // compute fast marching on tet to update t[3]
        ACG::TetFMM::fmm_tet( t, pts);
        
        // only clamp sign if vertex on mesh
        if( vvtovid < nv)
          fmm_dst[vvtovid] = std::max( 0., std::min( fmm_dst[vvtovid], t[3]));
        else
          fmm_dst[vvtovid] = std::min( fmm_dst[vvtovid], t[3]);
      }
      // enqueue and set CLOSE
      pq.insert( std::make_pair( fmm_dst[vvtovid], vvtovid));
      fmm_tag[vvtovid] = CLOSE;
    }
  }

  _dijtime = sw.stop()/1000.0;


  OpenMesh::VPropHandleT< Scalar > hulldist_prop;
  if( !_mesh.get_property_handle( hulldist_prop, "hulldist_prop"))
    _mesh.add_property( hulldist_prop, "hulldist_prop");
  Scalar mindist(FLT_MAX);
  Scalar maxdist(-FLT_MAX);
  int ninside(0);
  for( VIter v_it = _mesh.vertices_begin(); v_it != _mesh.vertices_end(); ++v_it)
  {
    int vid( v_it.handle().idx());

    //std::cerr << insidevertextag[vid] << " \t " << boundaryvertex[vid] << std::endl;
    _mesh.property( hulldist_prop, v_it) = fmm_dst[vid];
    mindist = std::min( mindist, fmm_dst[vid]);
    maxdist = std::max( maxdist, fmm_dst[vid]);
  }
}

////////////////////////////////////////////////////////

Vec3 EfficientSPC::closest_bsp_point( const Vec3& _pt, const MeshT& _bspmesh, OpenMeshTriangleBSPT< MeshT>* & _bsp)
{
  OpenMeshTriangleBSPT< MeshT >::Base::Base::NearestNeighbor nn = _bsp->nearest( _pt);

  TriMesh::CFVIter fv_it( _bspmesh.cfv_iter(nn.handle));
  Vec3 p0 = _bspmesh.point(fv_it);
  return ACG::Geometry::projectToPlane<Vec3>(p0, _bspmesh.normal(nn.handle), _pt);
}

////////////////////////////////////////////////////////

FH EfficientSPC::closest_bsp_fh( const Vec3& _pt, OpenMeshTriangleBSPT< MeshT>* & _bsp)
{
  OpenMeshTriangleBSPT< MeshT >::Base::Base::NearestNeighbor nn = _bsp->nearest( _pt);
  return nn.handle;
}

////////////////////////////////////////////////////////

void EfficientSPC::closest_bsp_ray( const Vec3& _pt, const Vec3& _ray, OpenMeshTriangleBSPT< MeshT>* & _bsp, std::vector< std::pair< FH, Scalar > >& _collisions)
{
  _collisions.clear();
  _collisions = _bsp->directionalRaycollision( _pt, _ray);
}

////////////////////////////////////////////////////////

Scalar EfficientSPC::tri_area( const FH& _fh, const MeshT& _mesh)
{
  std::vector< Scalar > v(3);
  int i(0);
  for( CFHIter fh_it( _mesh.cfh_iter( _fh)); fh_it; ++fh_it)
  {
    v[i] = ( _mesh.point( _mesh.to_vertex_handle( fh_it)) - _mesh.point( _mesh.from_vertex_handle( fh_it))).norm();
    i++;
  }
  std::sort( v.begin(), v.end());

  return 0.25*sqrt((v[0]+(v[1]+v[2]))*(v[2]-(v[0]-v[1]))*(v[2]+(v[0]-v[1]))*(v[0]+(v[1]-v[2])));
}

////////////////////////////////////////////////////////

Scalar EfficientSPC::tri_area( const std::vector< Vec3 >& _pts)
{
  std::vector< Scalar > v(3);
  for( int i = 0; i < 3; ++i)
  {
    v[i] = ( _pts[i] - _pts[(i+1)%3]).norm();
  }
  std::sort( v.begin(), v.end());

  return 0.25*sqrt(fabs((v[0]+(v[1]+v[2]))*(v[2]-(v[0]-v[1]))*(v[2]+(v[0]-v[1]))*(v[0]+(v[1]-v[2]))));
}

////////////////////////////////////////////////////////


} // namespace ACG

