/*
 * EfficientSPCPlugin -- Implementation
 *
 * Author:  Henrik Zimmer <henrik@zimmer.to>
 *
 * Version: $Revision: 1 $
 * Date:    $date$
 */

#include <Qt>
#include <QtGui>
#include <QSpacerItem>

#include "EfficientSPCPlugin.hh"

#include <iostream>
#include <iomanip>

#include <ACG/Utils/StopWatch.hh>
#include <ACG/Geometry/bsp/TriangleBSPT.hh>

#include <OpenFlipper/BasePlugin/PluginFunctions.hh>

#ifndef TETLIBRARY
#define TETLIBRARY 1
#endif

#undef min
#undef max

////////////////////////////////////////////////////////

void EfficientSPCPlugin::initializePlugin()
{
  tool_ = new EfficientSPCToolbar();
  
  QSize size(300,300);
  tool_->resize(size);
  
  // 1) compute convex hull
  connect(tool_->convex_pb, SIGNAL( clicked() ), this, SLOT( slot_convex_hull() ) );
  // 2) compute robust k-planes hull
  connect(tool_->k_planes_pb, SIGNAL( clicked() ), this, SLOT( slot_kplanes()));
  // 3) tetrahedralize 
  connect(tool_->mesh_ownstruct_pb, SIGNAL( clicked() ), this, SLOT( slot_tetrahedralize() ) );
  // 4) fmm
  connect(tool_->intersect_ownstruct_pb, SIGNAL( clicked() ), this, SLOT( slot_fmm() ) );

  connect(tool_->drawonering_sb, SIGNAL( valueChanged(int)), this, SLOT( slot_simple_draw_onering(int)));
  connect(tool_->drawmesh_cb, SIGNAL( stateChanged(int)), this, SLOT( slot_simple_draw_tetwireframe(int)));

  connect(tool_->save_pb, SIGNAL( clicked() ), this, SLOT( slot_save() ) );
  connect(tool_->load_pb, SIGNAL( clicked() ), this, SLOT( slot_load() ) );
  connect(tool_->moments_pb, SIGNAL( clicked() ), this, SLOT( slot_central_moments()));

  emit addToolbox( tr("EfficientSPCPlugin") , tool_ );
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::pluginsInitialized() 
{
//   emit addPickMode("Separator");
  momentsstr_.push_back( "mean          " );
  momentsstr_.push_back( "standard dev. " );
  momentsstr_.push_back( "variance      " );
  momentsstr_.push_back( "skewness      " );
  momentsstr_.push_back( "kurtosis      " );
}

////////////////////////////////////////////////////////

EfficientSPCPlugin::EfficientSPC*
EfficientSPCPlugin::
get_efficientspc_object( BaseObjectData* _object )
{
  // initialize PerObjectData if not done yet
  if (!_object->hasObjectData(pod_name()))
  {
    // get mesh object
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( _object )->mesh();

    // initialize per object data
    _object->setObjectData(pod_name(), new POD(*mesh));
  }
  
  // get feature lines object
  EfficientSPC* efficientspc = dynamic_cast<EfficientSPC*>(& (dynamic_cast< POD* >(_object->objectData(pod_name() )))->efficientspc());

  return efficientspc;
}

////////////////////////////////////////////////////////

EfficientSPCPlugin::POD*
EfficientSPCPlugin::
get_POD( BaseObjectData* _object )
{
  // initialize PerObjectData if not done yet
  if (!_object->hasObjectData(pod_name()))
  {
    // get mesh object
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( _object )->mesh();

    // initialize per object data
    _object->setObjectData(pod_name(), new POD(*mesh));
    // always try to load if there is something to load (prevents overwriting old results if only computing some new)
    //slot_load();
  }
  
  // get feature lines object
  return (dynamic_cast< POD* >(_object->objectData(pod_name() )));
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_convex_hull() 
{
  // iterate over all target triangle meshes
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      // Add new tet mesh
      int con_id;
      BaseObjectData *conbaseobj;
      TriMeshObject  *conobj;

      emit addEmptyObject(DATA_TRIANGLE_MESH, con_id);
      PluginFunctions::getObject(con_id, conbaseobj);
      conobj = PluginFunctions::triMeshObject(conbaseobj);
      conobj->setName("Convex Hull");
      conobj->materialNode()->disable_color_material();
      conobj->show();
      conobj->target(false);

      get_POD( *o_it)->convex_hull_id() = con_id;
      Scalar& contime( get_POD( *o_it)->contime());
      contime = ACG::EfficientSPC::do_convex_hull( *mesh, *(conobj->mesh()));
      std::cerr << " --- Convex Hull Result --- " << std::endl;
      std::cerr << " #meshvertices = " << mesh->n_vertices() << std::endl;
      std::cerr << " #hullvertices = " << conobj->mesh()->n_vertices() << std::endl;
      std::cerr << " hull time     = " << contime << std::endl;
      std::cerr << " -------------------------- " << std::endl;

      emit updatedObject( con_id, UPDATE_ALL);
    }
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_kplanes()
{
  // iterate over all target triangle meshes
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      // get convex hull and plc
      int con_id( get_POD( *o_it)->convex_hull_id());

      TriMeshObject  *conobj;
      BaseObjectData *conbaseobj;
      PluginFunctions::getObject(con_id, conbaseobj);
      conobj = PluginFunctions::triMeshObject(conbaseobj);

      if( conobj == NULL )
      {
        std::cerr << __FUNCTION__ << " convex hull not yet computed!" << std::endl;
        return;
      }
      Scalar avgel( get_POD( *o_it)->avgel());
      TriMesh& conmesh( *(conobj->mesh()));

      OpenMeshTriangleBSPT< TriMesh > *& chbsp( get_POD( conbaseobj)->bsp());

      if( chbsp == NULL)
      {
        std::cerr << __FUNCTION__ << " requesting convex hull BSP " << std::endl;
        chbsp = conobj->requestTriangleBsp();
      }

      ACG::StopWatch sw;
      sw.start();

      // Add new polymesh
      int& kplanes_id(get_POD(*o_it)->kplanesmesh_id());
      BaseObjectData *kplanesbaseobj;
      PolyMeshObject  *kplanesobj;

      emit addEmptyObject(DATA_POLY_MESH, kplanes_id);
      PluginFunctions::getObject(kplanes_id, kplanesbaseobj);
      kplanesobj = PluginFunctions::polyMeshObject(kplanesbaseobj);
      kplanesobj->setName("k-planes");
      kplanesobj->materialNode()->disable_color_material();
      kplanesobj->show();
      kplanesobj->target(false);
      PMeshT& kplanesmesh( *(kplanesobj->mesh()));

      int k( tool_->kplanes_sb->value());

      std::vector< Vec3 > normals;
      normals.reserve( conmesh.n_faces() + 6 + 22);

      normals.push_back( Vec3( 1., 0., 0.));
      normals.push_back( Vec3( 0., 1., 0.));
      normals.push_back( Vec3( 0., 0., 1.));
      normals.push_back( Vec3( -1., 0., 0.));
      normals.push_back( Vec3( 0., -1., 0.));
      normals.push_back( Vec3( 0., 0., -1.));
      Vec3 tmpn( 1,1,0);
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 1,-1,0);
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 1,0,1);
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 1,0,-1);
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 0,1,1);
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 0,-1,1);
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);

      tmpn = Vec3( 1,1,1); // -1 -1 -1
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( -1,1,1); // 1 -1 -1
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 1,-1,1); // -1 1 -1
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);
      tmpn = Vec3( 1,1,-1); // -1 -1 1
      tmpn.normalize();
      normals.push_back( tmpn);
      normals.push_back( -tmpn);

      ACG::SceneGraph::BaseNode* bn = PluginFunctions::getRootNode();
      ACG::SceneGraph::MaterialNode* matnin = new ACG::SceneGraph::MaterialNode( bn, "MatnodeIN");
      (*o_it)->addAdditionalNode(matnin, name(), "MatnodeIN");
      ACG::SceneGraph::PointNode* pnins = new ACG::SceneGraph::PointNode(matnin, "InsidePts");
      (*o_it)->addAdditionalNode(pnins, name(), "InsidePts");

      std::vector< Vec3 > samples;
      samples.reserve( k + 30);
      // create extra 26 + k samples on sphere
      for( int i = 0; i < k; ++i)
      {
        Scalar z(rand()/((Scalar)RAND_MAX)*2.0 - 1.0);
        Scalar phi(rand()/((Scalar)RAND_MAX)*2.0*M_PI);
        Scalar x( sqrt(std::max(0.0,1.0-z*z)));
        Scalar y( x * sin(phi));
        x *= cos(phi);
        Vec3 sample( x,y,z);
        samples.push_back( sample);
      }

      // add the 26-DOP samples on the sphere plus extra directions added to normals
      for( size_t j = 0; j < normals.size(); ++j)
        samples.push_back( normals[j]);

      size_t nclusters( samples.size());

      normals.clear();
      normals.resize( conmesh.n_faces());
      size_t nnormals( normals.size());
      std::vector< Scalar > areas( nnormals, 0.0);

      // add rest of normals
      for( FIter f_it = conmesh.faces_begin(); f_it != conmesh.faces_end(); ++f_it)
      {
        int fid( f_it.handle().idx());
        areas[fid] = fabs(ACG::EfficientSPC::tri_area( f_it, conmesh));
        normals[fid] =  -conmesh.normal( f_it.handle());
      }

      ////! k-means clustering of samples on sphere (weighted by convex hull triangle areas)
      std::vector< std::pair< Scalar, int > > closestclustertonormal(nnormals, std::pair<Scalar,int>(FLT_MAX, -1));
      std::vector< Vec3 > meanofcluster(nclusters);
      std::vector< size_t > normalsincluster( nclusters, 0);
      std::vector< size_t > lastnormalsincluster( nclusters, 0);
      bool converged( false);
      int infsave(1000); // avoid \inf iterations
      while( !converged && infsave > 0)
      {
        infsave--;
        lastnormalsincluster.assign( normalsincluster.begin(), normalsincluster.end());
        closestclustertonormal.clear();
        closestclustertonormal.resize( nnormals, std::pair<Scalar,int>(FLT_MAX, -1));
        // compute all k*normals distances
        for( size_t n = 0; n < nnormals; ++n)
        {
          for( size_t j = 0; j < nclusters; ++j)
          {
            Scalar dist( fabs(areas[n] * acos( std::min( 1.0, std::max( -1.0, normals[n] | samples[j])))));
            if( dist < closestclustertonormal[n].first)
            {
              closestclustertonormal[n].first = dist;
              closestclustertonormal[n].second = j;
            }
          }
        }

        // compute means
        meanofcluster.clear();
        meanofcluster.resize( nclusters, Vec3(0.,0.,0.));
        normalsincluster.clear();
        normalsincluster.resize( nclusters, 0);
        for( size_t n = 0; n < nnormals; ++n)
        {
          meanofcluster[closestclustertonormal[n].second] += areas[n]*normals[n];
          normalsincluster[closestclustertonormal[n].second]++;
        }
        bool finished(true);
        for( size_t c = 0; c < nclusters; ++c)
        {
          if( normalsincluster[c] != lastnormalsincluster[c])
            finished = false;

          if( normalsincluster[c] != 0)
            samples[c] = meanofcluster[c].normalize();/// Scalar(normalsincluster[c])).normalize();
        }
          //else
          //  samples[c] = samples[c];
        if( finished )
          converged = true;
      }

      if( converged )
        std::cerr << " K-MEANS converged! iters " << (1000-infsave)<< std::endl;
      else
        std::cerr << " K-MEANS did \t NOT \t CONVERGE!" << std::endl;

      ////! remove too similar clusters
      int nreduced(0);
      bool reducedone(false);
      while( !reducedone)
      {
        reducedone = true;
        for( size_t i = 0; i < samples.size() && !reducedone; ++i)
        {
          for( size_t j = 0; j < samples.size() && !reducedone; ++j)
          {
            if( i != j)
            {
              if( (acos(samples[j]|samples[i])*180./M_PI) < 0.5)
              {
                samples.erase( samples.begin() + i);
                reducedone = false;
                nreduced++;
              }
            }
          }
        }
      }

      if( nreduced)
        std::cerr << " ... simplifed k-DOP by " << nreduced << " planes " << std::endl;
      

      // find inside point (cog of hull)
      Vec3 insidept( 0., 0., 0.);
      for( VIter v_it = conmesh.vertices_begin(); v_it != conmesh.vertices_end(); ++v_it)
      {
        insidept += conmesh.point( v_it);
      }
      insidept /= Scalar( conmesh.n_vertices());

      // find closest point on convex hull from all planes
      std::vector< Scalar > maxdistfrominsidept( nclusters, -FLT_MAX);
      for( size_t c = 0; c < nclusters; ++c)
      {
        for( VIter v_it = conmesh.vertices_begin(); v_it != conmesh.vertices_end(); ++v_it)
        {
          Scalar dist( samples[c] | (conmesh.point(v_it)-insidept));
          if( dist > maxdistfrominsidept[c])
          {
            maxdistfrominsidept[c] = dist;
          }
        }
      }

      // offset distances
      for( size_t c = 0; c < nclusters; ++c)
      {
        maxdistfrominsidept[c] += avgel;
        //std::cerr << maxdistfrominsidept[c] << std::endl;
      }

      // dual points
      std::vector< Point_3 > dualpoints( nclusters);
      for( size_t c = 0; c < nclusters; ++c)
      {
        Vec3 dualpt(-samples[c]/maxdistfrominsidept[c]);

        dualpoints[c] = Point_3(dualpt[0], dualpt[1], dualpt[2]);
      }

      // define object to hold convex hull
      CGAL::Object ch_object;

      // compute convex hull
      CGAL::convex_hull_3(dualpoints.begin(), dualpoints.end(), ch_object);
      Polyhedron_3 convexhull;

      // determine what kind of object it is
      if (CGAL::object_cast<Segment_3>(&ch_object) )
      {
        std::cerr << "convex hull is a segment " << std::endl;
        return;
      }
      else if (CGAL::object_cast<Polyhedron_3>(&ch_object) )
      {
        std::cerr << "convex hull is a polyhedron " << std::endl;
        convexhull = CGAL::object_cast<Polyhedron_3>(ch_object);
      }
      else
      {
        std::cerr << "convex hull error!" << std::endl;
        return;
      }

      std::transform( convexhull.facets_begin(), convexhull.facets_end(), convexhull.planes_begin(),Plane_from_facet());

      // compute normals of convex hull
      size_t sizef = convexhull.size_of_facets();
      std::cerr << " sizef " << sizef << std::endl;

      PolyFIter faceit = convexhull.facets_begin();
      PolyFIter facend = convexhull.facets_end();

      std::map< PolyFIter, VH > vertmap;
      // hull face normal = primal points
      for(; faceit != facend; ++faceit )
      {
        Plane_3 fplane = faceit->plane();
        Vertex_handle vh = faceit->halfedge()->vertex();
        Vec3 primalpt( fplane.a(), fplane.b(), fplane.c());
        primalpt /= fplane.d();
        vertmap[faceit] = kplanesmesh.add_vertex( primalpt + insidept);
      }

      // iterate over hull vertices and incident faces and build primal mesh
      PolyVIter vertit = convexhull.vertices_begin();
      PolyVIter vertnd = convexhull.vertices_end();

      for(; vertit != vertnd; ++vertit)
      {
        PolyVHIter vhit( vertit->vertex_begin());
        PolyVHIter vhnd( vertit->vertex_begin());
        std::vector< VH > face;
        do
        {
          Polyhedron_3::Facet_handle curfh( vhit->facet());
          face.push_back( vertmap[curfh]);
          ++vhit;
        } while( vhit != vhnd);
        kplanesmesh.add_face( face);
      }


      ////! simplify topology a bit (collapse very short edges)
      int ncollapsed(0);
      for( EdgeIter e_it = kplanesmesh.edges_begin(); e_it != kplanesmesh.edges_end(); ++e_it)
        if( kplanesmesh.calc_edge_length( e_it.handle()) < avgel * 0.1)
        {
          kplanesmesh.collapse( kplanesmesh.halfedge_handle( e_it.handle(), 0));
          ncollapsed++;
        }

      if( ncollapsed)
        std::cerr << " collapsed " << ncollapsed << " edges .." << std::endl;
      kplanesmesh.garbage_collection();
      kplanesmesh.update_normals();

      get_POD( *o_it)->remtime() = sw.stop()/1000.0;
      std::cerr << " Robust Outer Hull took " << get_POD(*o_it)->remtime() << std::endl;
      
      emit updatedObject( kplanes_id, UPDATE_ALL);
    }
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_tetrahedralize()
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      ACG::LeanTetMesh& tetmesh( get_POD( *o_it)->leanmesh());

      // get convex hull
      int con_id( get_POD( *o_it)->convex_hull_id());

      TriMeshObject  *conobj;
      BaseObjectData *conbaseobj;
      PluginFunctions::getObject(con_id, conbaseobj);

      if( conbaseobj == NULL )
      {
        std::cerr << __FUNCTION__ << " convex hull not yet computed!" << std::endl;
        return;
      }
      conobj = PluginFunctions::triMeshObject(conbaseobj);
      Scalar avgel( get_POD( *o_it)->avgel());

      OpenMeshTriangleBSPT< TriMesh > *& chbsp( get_POD( conbaseobj)->bsp());

      if( chbsp == NULL)
      {
        std::cerr << __FUNCTION__ << " requesting convex hull BSP " << std::endl;
        chbsp = conobj->requestTriangleBsp();
      }

      int kplanes_id(get_POD(*o_it)->kplanesmesh_id());
      BaseObjectData *kplanesbaseobj;
      PluginFunctions::getObject(kplanes_id, kplanesbaseobj);
      if( kplanesbaseobj == NULL )
      {
        std::cerr << __FUNCTION__ << " convex hull not yet computed!" << std::endl;
        return;
      }
      PolyMeshObject  *kplanesobj( PluginFunctions::polyMeshObject(kplanesbaseobj));
      PMeshT& kplanesmesh( *(kplanesobj->mesh()));


      Scalar& cdttime( get_POD( *o_it)->cdttime());
      cdttime = ACG::EfficientSPC::tetrahedralize( *mesh, kplanesmesh, avgel, tetmesh, tool_->edgescale_dsb->value(), tool_->aspect_dsb->value());
      std::cerr << " CDT time      = " << cdttime << std::endl;


      emit updatedObject( o_it->id(), UPDATE_ALL);
    }
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_fmm()
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      ACG::LeanTetMesh& tetmesh( get_POD( *o_it)->leanmesh());

      // get convex hull and plc
      int con_id( get_POD( *o_it)->convex_hull_id());

      TriMeshObject  *conobj;
      BaseObjectData *conbaseobj;
      PluginFunctions::getObject(con_id, conbaseobj);
      conobj = PluginFunctions::triMeshObject(conbaseobj);

      if( conobj == NULL )
      {
        std::cerr << __FUNCTION__ << " convex hull not yet computed!" << std::endl;
        return;
      }
      Scalar avgel( get_POD( *o_it)->avgel());


      OpenMeshTriangleBSPT< TriMesh > *& chbsp( get_POD( conbaseobj)->bsp());

      if( chbsp == NULL)
      {
        std::cerr << __FUNCTION__ << " requesting convex hull BSP " << std::endl;
        chbsp = conobj->requestTriangleBsp();
      }


      Scalar& inttime( get_POD( *o_it)->inttime());
      Scalar& dijtime( get_POD( *o_it)->dijtime());

      ACG::EfficientSPC::fmm( tetmesh, *(conobj->mesh()), chbsp, avgel, *mesh, inttime, dijtime);
      std::cerr << " --- Fast Marching Result --- " << std::endl;
      std::cerr << " intersection time   = " << inttime << std::endl;
      std::cerr << " marching time       = " << dijtime << std::endl;
      std::cerr << " ---------------------------- " << std::endl;

    }
    emit updatedObject( o_it->id(), UPDATE_SELECTION);
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_simple_draw_onering( int _i)
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      ACG::LeanTetMesh& tetmesh( get_POD( *o_it)->leanmesh());

      ACG::SceneGraph::BaseNode* bn = PluginFunctions::getRootNode(); 

      ACG::SceneGraph::LineNode*& ln( get_POD(*o_it)->ln_);
      ACG::SceneGraph::PointNode*& pn( get_POD(*o_it)->pn_);
      ACG::SceneGraph::MaterialNode*& mn( get_POD(*o_it)->mn_);

      if( ln == NULL)
      {
        ln = new ACG::SceneGraph::LineNode(ACG::SceneGraph::LineNode::LineSegmentsMode, bn, "myLineNode");
        (*o_it)->addAdditionalNode(ln, name(), "myLineNode");
      }
      if( mn == NULL)
      {
        mn = new ACG::SceneGraph::MaterialNode( bn, "myMatNode");
        (*o_it)->addAdditionalNode(mn, name(), "myMatNode");
        mn->set_point_size( 3);
      }
      if( pn == NULL)
      {
        pn = new ACG::SceneGraph::PointNode(mn, "myPointNode");
        (*o_it)->addAdditionalNode(pn, name(), "myPointNode");
      }


      pn->show();
      pn->drawMode( ACG::SceneGraph::DrawModes::POINTS_COLORED);

      ln->clear();
      ln->set_line_width(3);

      std::set<int>& vv(tetmesh.vv[_i]);
      for( std::set<int>::iterator it( vv.begin()); it != vv.end(); ++it)
      {
        ln->add_line( tetmesh.pts[_i], tetmesh.pts[*it]);
        ln->add_color( ACG::Vec4f(1,0,0,1));
      }
    }
    emit updatedObject( o_it->id(), UPDATE_SELECTION);
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_simple_draw_tetwireframe(int state)
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      ACG::LeanTetMesh& tetmesh( get_POD( *o_it)->leanmesh());

      ACG::SceneGraph::BaseNode* bn = PluginFunctions::getRootNode(); 

      ACG::SceneGraph::LineNode*& ln( get_POD(*o_it)->ln2_);

      if( ln == NULL)
      {
        ln = new ACG::SceneGraph::LineNode(ACG::SceneGraph::LineNode::LineSegmentsMode, bn, "myLineNode2");
        (*o_it)->addAdditionalNode(ln, name(), "myLineNode2");
      }
      ln->clear();
      ln->set_line_width(1);
      if( state)
      {
        tetmesh.build_edges();
        for( size_t i = 0; i < tetmesh.ne; ++i)
        {
          ln->add_line( tetmesh.pts[tetmesh.edges[i].id0], tetmesh.pts[tetmesh.edges[i].id1]);
          ln->add_color( ACG::Vec4f(1,1,0,0.4));
        }
      }
    }
    emit updatedObject( o_it->id(), UPDATE_SELECTION);
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_save( )
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      OpenMesh::VPropHandleT< double > hulldist_prop;
      if( !mesh->get_property_handle( hulldist_prop, "hulldist_prop"))
      {
        std::cerr << __FUNCTION__ << " NO HULL DIST COMPUTED!" << std::endl;
        return;
      }

      // get a file name
      QString def_fileName((QDir::homePath ()).append("/projects/ConvexityMeasure/data/"));
      //def_fileName.append(o_it->name()).append(".con");
      //QString fileName( def_fileName);
      QString fileName = QFileDialog::getSaveFileName(tool_, tr("Save concavity stuff"), def_fileName, tr("Concavity Data (*.con)"));

      //QString fileName( def_fileName);
      std::ofstream savefile(fileName.toStdString().c_str());
      savefile << "VALUES" << std::endl;
      savefile << std::setprecision(8);
      for( VIter v_it = mesh->vertices_begin(); v_it != mesh->vertices_end(); ++v_it)
      {
        savefile << mesh->property( hulldist_prop, v_it) << std::endl;
      }

      Scalar normalization( get_POD( *o_it)->sqrtarea());
      if( normalization != 1.0)
      {
        savefile << "NORMALIZATION" << std::endl;
        savefile << normalization << std::endl;
      }

      // save central moments
      Scalar mean( get_POD( *o_it)->mean());
      Scalar std( get_POD( *o_it)->std());
      Scalar var( get_POD( *o_it)->var());
      Scalar skew( get_POD( *o_it)->skew());
      Scalar kurt( get_POD( *o_it)->kurt());
      if( mean != -1.0)
      {
        savefile << "MOMENTS" << std::endl;
        savefile << mean << std::endl; 
        savefile << std << std::endl; 
        savefile << var << std::endl; 
        savefile << skew << std::endl; 
        savefile << kurt << std::endl; 
      }

      Scalar remtime( get_POD( *o_it)->remtime());
      Scalar contime( get_POD( *o_it)->contime());
      Scalar cdttime( get_POD( *o_it)->cdttime());
      Scalar inttime( get_POD( *o_it)->inttime());
      Scalar dijtime( get_POD( *o_it)->dijtime());
      if( remtime != -1.0)
      {
        savefile << "REMTIME" << std::endl;
        savefile << remtime << std::endl;
      }
      if( contime != -1.0)
      {
        savefile << "CONTIME" << std::endl;
        savefile << contime << std::endl;
      }
      if( cdttime != -1.0)
      {
        savefile << "CDTTIME" << std::endl;
        savefile << cdttime << std::endl;
      }
      if( inttime != -1.0)
      {
        savefile << "INTTIME" << std::endl;
        savefile << inttime << std::endl;
      }
      if( dijtime != -1.0)
      {
        savefile << "DIJTIME" << std::endl;
        savefile << dijtime << std::endl;
      }

      savefile.close();
    }
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_load( )
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {
    // get pointer to mesh
    TriMesh* mesh = dynamic_cast< TriMeshObject* >( *o_it )->mesh();
    if ( mesh )
    {
      OpenMesh::VPropHandleT< double > hulldist_prop;
      if( !mesh->get_property_handle( hulldist_prop, "hulldist_prop"))
      {
        mesh->add_property( hulldist_prop, "hulldist_prop");
      }

      // get a file name
      QString def_fileName((QDir::homePath ()).append("/projects/ConvexityMeasure/data/"));
      //def_fileName.append(o_it->name()).append(".con");
      QString fileName = QFileDialog::getOpenFileName(tool_, tr("Load concavity stuff"), def_fileName, tr("Concavity Data (*.con)"));
      //QString fileName( def_fileName);

      std::ifstream loadfile(fileName.toStdString().c_str());
      double mindist(FLT_MAX);
      double maxdist(-FLT_MAX);

      std::string line;
      while (std::getline(loadfile,line)) 
      {
        std::string s;
        //line >> s;
        //std::istringstream(line) >> s;
        if( line == "VALUES" )
        {
          for( VIter v_it = mesh->vertices_begin(); v_it != mesh->vertices_end(); ++v_it)
          {
            double val(0.0);
            loadfile >> val;
            mindist = std::min( mindist, val);
            maxdist = std::max( maxdist, val);
            mesh->property( hulldist_prop, v_it) = val;
          }
        }
        if( line == "NORMALPROJTOHULLVALUES")
        {
          for( VIter v_it = mesh->vertices_begin(); v_it != mesh->vertices_end(); ++v_it)
          {
            double val(0.0);
            loadfile >> val;
            // ignore (DEPRECATED)
          }
        }
        if( line == "NORMALIZATION" )
        {
          double val(-1.);
          loadfile >> val;
          get_POD( *o_it)->sqrtarea() = val;
        }
        if( line == "MOMENTS")
        {
          Scalar& mean( get_POD( *o_it)->mean());
          Scalar& std( get_POD( *o_it)->std());
          Scalar& var( get_POD( *o_it)->var());
          Scalar& skew( get_POD( *o_it)->skew());
          Scalar& kurt( get_POD( *o_it)->kurt());
          loadfile >> mean;
          loadfile >> std;
          loadfile >> var;
          loadfile >> skew;
          loadfile >> kurt;
        }

        Scalar& remtime( get_POD( *o_it)->remtime());
        Scalar& contime( get_POD( *o_it)->contime());
        Scalar& cdttime( get_POD( *o_it)->cdttime());
        Scalar& inttime( get_POD( *o_it)->inttime());
        Scalar& dijtime( get_POD( *o_it)->dijtime());
        if( line == "REMTIME")
          loadfile >> remtime;
        if( line == "CONTIME")
          loadfile >> contime;
        if( line == "CDTTIME")
          loadfile >> cdttime;
        if( line == "INTTIME")
          loadfile >> inttime;
        if( line == "DIJTIME")
          loadfile >> dijtime;
      }
      loadfile.close();
      ACG::ColorCoder cc(mindist, maxdist, false );
      for( VIter v_it = mesh->vertices_begin(); v_it != mesh->vertices_end(); ++v_it)
      {
        int vid( v_it.handle().idx());
        mesh->set_color( v_it, cc.color_float4(mesh->property( hulldist_prop, v_it)));
      }
      std::cerr << __FUNCTION__ << " mindist " << mindist << " maxdist " << maxdist << std::endl;
      emit updatedObject( (*o_it)->id(), UPDATE_COLOR);
    }
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::slot_central_moments()
{
  for ( PluginFunctions::ObjectIterator o_it(PluginFunctions::TARGET_OBJECTS,DATA_TRIANGLE_MESH) ; o_it != PluginFunctions::objectsEnd(); ++o_it) 
  {

    std::vector< Scalar > moms;
    compute_central_moments( *o_it, moms);
    std::cerr << " --- Central Moments ---" << std::endl;
    for( size_t i = 0; i < moms.size(); ++i)
      std::cerr << momentsstr_[i] << moms[i] << std::endl;
    std::cerr << " ----------------------- " << std::endl;
  }
}

////////////////////////////////////////////////////////

void EfficientSPCPlugin::compute_central_moments( BaseObjectData* _object, std::vector< Scalar >& _moms)
{
  TriMesh* mesh = dynamic_cast< TriMeshObject* >( _object )->mesh();
  if ( mesh )
  {
    OpenMesh::VPropHandleT< double > hulldist_prop;
    if( !mesh->get_property_handle( hulldist_prop, "hulldist_prop"))
    {
      std::cerr << __FUNCTION__ << " need property for computing moments" << std::endl;
      return;
    }

    // integrate over surface normalize with (and clamp to) longest line through CH or longest shortest path on mesh
    Scalar normalization( get_POD( _object)->sqrtarea());

    std::vector< Scalar > values( mesh->n_vertices(), 0.);

    // normalize (and clamp) 
    for( VIter v_it = mesh->vertices_begin(); v_it != mesh->vertices_end(); ++v_it)
    {
      double val(mesh->property( hulldist_prop, v_it));
      values[v_it.handle().idx()] = std::min( val/normalization, 1.0);
    }

    // integrate mean
    Scalar totarea(0.0);
    Scalar meanintegral(0.0);
    for( FIter f_it = mesh->faces_begin(); f_it != mesh->faces_end(); ++f_it)
    {
      Scalar area( ACG::EfficientSPC::tri_area( f_it, *mesh));
      Scalar faceint( 0.0);
      FVIter fv_it( mesh->fv_iter(f_it));
      for( ; fv_it; ++fv_it)
      {
        int vid( fv_it.handle().idx());
        faceint += values[vid]*1.0/3.0;
      }
      faceint *= area;
      meanintegral += faceint;
      totarea += area;
    }
    Scalar& mean( get_POD( _object)->mean());
    mean = meanintegral/totarea;
    //std::cerr << " totarea " << totarea << std::endl;
    //std::cerr << " meanintegral " << meanintegral << " mean " <<mean << std::endl;

    // integrate variance
    Scalar varintegral(0.0);
    for( FIter f_it = mesh->faces_begin(); f_it != mesh->faces_end(); ++f_it)
    {
      Scalar area( ACG::EfficientSPC::tri_area( f_it, *mesh));
      Scalar faceint( 0.0);
      FVIter fv_it( mesh->fv_iter(f_it));
      for( ; fv_it; ++fv_it)
      {
        int vid( fv_it.handle().idx());
        faceint += (values[vid]-mean)*(values[vid]-mean)*1.0/3.0;
      }
      faceint *= area;
      varintegral += faceint;
    }
    Scalar& var( get_POD( _object)->var());
    var = varintegral/totarea;
    Scalar& std( get_POD( _object)->std());
    std = sqrt(var);
    //std::cerr << " \t\tstandard deviation " << std << std::endl;
    //std::cerr << " varintegral " << varintegral << " var " << var << std::endl;

    // skewness integral
    Scalar skewintegral(0.0);
    for( FIter f_it = mesh->faces_begin(); f_it != mesh->faces_end(); ++f_it)
    {
      Scalar area( ACG::EfficientSPC::tri_area( f_it, *mesh));
      Scalar faceint( 0.0);
      FVIter fv_it( mesh->fv_iter(f_it));
      for( ; fv_it; ++fv_it)
      {
        int vid( fv_it.handle().idx());
        faceint += (values[vid]-mean)*(values[vid]-mean)*(values[vid]-mean)*1.0/3.0;
      }
      faceint *= area;
      skewintegral += faceint;
    }
    Scalar& skew( get_POD( _object)->skew());
    skew = skewintegral/totarea;
    Scalar unnormskew(skew);
    skew /= sqrt(var*var*var);
    //std::cerr << " skewintegral " << skewintegral << " skew " << skew << std::endl;
    //std::cerr << " \t\t unnormalized skew " << unnormskew << std::endl;

    // kurtosis integral
    Scalar kurtintegral(0.0);
    for( FIter f_it = mesh->faces_begin(); f_it != mesh->faces_end(); ++f_it)
    {
      Scalar area( ACG::EfficientSPC::tri_area( f_it, *mesh));
      Scalar faceint( 0.0);
      FVIter fv_it( mesh->fv_iter(f_it));
      for( ; fv_it; ++fv_it)
      {
        int vid( fv_it.handle().idx());
        faceint += (values[vid]-mean)*(values[vid]-mean)*(values[vid]-mean)*(values[vid]-mean)*1.0/3.0;
      }
      faceint *= area;
      kurtintegral += faceint;
    }
    Scalar& kurt( get_POD( _object)->kurt());
    kurt = (kurtintegral/totarea);
    Scalar unnormkurz(kurt);
    kurt /= var*var;
    //std::cerr << " kurtintegral " << kurtintegral << " kurt " << kurt << std::endl;
    //std::cerr << " \t\t unnormalized kurt " << unnormkurz << std::endl;

    _moms.clear();
    _moms.push_back( mean);
    _moms.push_back( std);
    _moms.push_back( var);
    _moms.push_back( skew);
    _moms.push_back( kurt);
  }
}

////////////////////////////////////////////////////////

Q_EXPORT_PLUGIN2( efficientspcplugin , EfficientSPCPlugin );

