/*
 * EfficientSPCPlugin -- Header
 *
 * Author:  Henrik Zimmer <henrik@zimmer.to>
 *
 * Version: $Revision: 1 $
 * Date:    $date$
 */


#ifndef EFFICIENTSPCPLUGIN_HH
#define EFFICIENTSPCPLUGIN_HH

#include <QObject>
#include <QMenuBar>
#include <QSpinBox>

#include <OpenFlipper/common/Types.hh>
#include <OpenFlipper/BasePlugin/BaseInterface.hh>
#include <OpenFlipper/BasePlugin/ToolboxInterface.hh>
#include <OpenFlipper/BasePlugin/KeyInterface.hh>
#include <OpenFlipper/BasePlugin/MouseInterface.hh>
#include <OpenFlipper/BasePlugin/PickingInterface.hh>
#include <OpenFlipper/BasePlugin/ScriptInterface.hh>
#include <OpenFlipper/BasePlugin/LoadSaveInterface.hh>

#include <ACG/QtWidgets/QtExaminerViewer.hh>

#include <ObjectTypes/TriangleMesh/TriangleMesh.hh>
#include <ObjectTypes/PolyMesh/PolyMesh.hh>

#include <ACG/Scenegraph/LineNode.hh>
#include <ACG/Scenegraph/PointNode.hh>

#include "EfficientSPCToolbar.hh"
#include "EfficientSPCPerObjectData.hh"
#include "EfficientSPC.hh"

#include <tetgen.h>
#include <ACG/Utils/ColorCoder.hh>

#include <ACG/Geometry/bsp/TriangleBSPT.hh> 


#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Exact_predicates_exact_constructions_kernel.h>
#include <CGAL/point_generators_3.h>
#include <CGAL/algorithm.h>
#include <CGAL/Convex_hull_traits_3.h>
#include <CGAL/convex_hull_3.h>

typedef CGAL::Exact_predicates_inexact_constructions_kernel  K;
typedef CGAL::Convex_hull_traits_3<K>             Traits;
typedef Traits::Polyhedron_3                      Polyhedron_3;
typedef CGAL::Segment_3<K>                              Segment_3;

typedef Polyhedron_3::Halfedge_around_facet_circulator PolyFHIter;
typedef Polyhedron_3::Facet_iterator PolyFIter;
typedef Polyhedron_3::Halfedge_around_vertex_circulator PolyVHIter;
typedef Polyhedron_3::Vertex_handle Vertex_handle;
typedef Polyhedron_3::Vertex_iterator PolyVIter;

typedef K::Point_3                                Point_3;
typedef Polyhedron_3::Facet::Plane_3  Plane_3;
typedef CGAL::Creator_uniform_3<double, Point_3>  PointCreator;

typedef ACG::VectorT<double, 3> Vec3;


struct Plane_from_facet {
  Polyhedron_3::Plane_3 operator()(Polyhedron_3::Facet& f) {
    Polyhedron_3::Halfedge_handle h = f.halfedge();
    return Polyhedron_3::Plane_3( h->vertex()->point(),
        h->next()->vertex()->point(),
        h->opposite()->vertex()->point());
  }
};


class EfficientSPCPlugin : public QObject, BaseInterface, ToolboxInterface, KeyInterface, ScriptInterface, MouseInterface, PickingInterface, LoadSaveInterface
{
  Q_OBJECT
  Q_INTERFACES(BaseInterface)
  Q_INTERFACES(ToolboxInterface)
  Q_INTERFACES(KeyInterface)
  Q_INTERFACES(ScriptInterface)
  Q_INTERFACES(MouseInterface)
  Q_INTERFACES(PickingInterface)
  Q_INTERFACES(LoadSaveInterface)


  // typedef for easy access
  typedef ACG::EfficientSPC EfficientSPC;  // SPC Object
  typedef EfficientSPCPerObjectData   POD; // Per Object Data

signals:
  void updateView();
  void updatedObject(int /*_identifier*/, const UpdateType /*_type*/);
  
  // LoadSaveInterface
  void addEmptyObject( DataType _type, int& _id);
  void deleteObject( int _id );
  void load(QString _filename, DataType _type, int& _id);
  
  // ToolboxInterface
  void addToolbox( QString _name  , QWidget* _widget );   


private slots:

  // initialization functions
  void initializePlugin();
  void pluginsInitialized();

  // convex hull
  void slot_convex_hull();
  // find compromise between convex hull (needs remeshing) and bb (to much free space) k-means cluster the normals
  void slot_kplanes();
  // tetrahedralize
  void slot_tetrahedralize();
  // fmm
  void slot_fmm();

  // load/save
  void slot_save();
  void slot_load();

  void slot_central_moments();
  void compute_central_moments( BaseObjectData* _object, std::vector< Scalar >& _moms);

  // very simple visualization (use www.openvolumemesh.org) for more/better visualization options
  void slot_simple_draw_onering(int);
  void slot_simple_draw_tetwireframe(int state=0);

public :

  ~EfficientSPCPlugin() {};

  QString name() { return (QString("EfficientSPC")); };
  QString description( ) { return (QString("Computes the EfficientSPC of the the active Mesh")); }; 

private :

  // return name of per object data
  const char * pod_name() { return "EFFICIENTSPC_PER_OBJECT_DATA";}

  // get EfficientSPC object for a given object
  EfficientSPC* get_efficientspc_object( BaseObjectData* _object );
  POD* get_POD( BaseObjectData* _object );
  

private :
  /// Widget for Toolbox
  EfficientSPCToolbar* tool_;
  std::vector< std::string > momentsstr_;
};


#endif // EFFICIENTSPCPLUGIN_HH defined

